from abc import ABC, abstractmethod
from .rate_limit import RateLimit
from .exceptions import RateLimitedException


class Store(ABC):
    key: str

    @abstractmethod
    def get_tat(self): ...

    @abstractmethod
    def set_tat(self, value: str): ...

    @abstractmethod
    def get_time(self): ...

    def update(self, limit: RateLimit):
        now = self.get_time()
        tat = max(self.get_tat(self.key), now)
        separation = tat - now
        max_interval = limit.period - limit.inverse
        if separation > max_interval:
            raise RateLimitedException()
        else:
            new_tat = max(tat, now) + limit.inverse
            self.set_tat(self.key, new_tat)



from .timestamp_typedef import ts

class RateLimit:
    def __init__(self, count: int = 2, period: ts = 1) -> None:
        '''
        count: number of requests in time period
        period: time period (seconds) 
        '''
        self.count = count
        self.period = period

    @property
    def inverse(self) -> float:
        return self.period / self.count




from time import sleep
from rerate import RateLimitDecorator, RateLimitedException

limit = RateLimitDecorator(limit=2, period=1, store_option='REDIS', key='my_api_key')

@limit
def example():
    sleep(0.4)
    print('hi')

for _ in range(10000):
    try:
        example()
    except RateLimitedException as e:
        print(e)



from typing import Literal, Optional
from .abstract_store import Store
from .dictionary_store import DictionaryStore
from .redis_store import RedisStore
from .rate_limit import RateLimit
from functools import wraps
from .exceptions import BadStoreOption

StoreOption = Literal['REDIS', 'DICTIONARY']

key = None
store_option = 'DICTIONARY'
host = 'localhost' 
port = 6379 

class RateLimitDecorator:
    store: Store
    limit: RateLimit

    def __init__(self, count: int = 15, period: int = 900, store_option: StoreOption = store_option, host: Optional[str] = host, port: Optional[int] = port, key: Optional[str] = key):
        limit = RateLimit(count=count, period=period)
        match store_option:
            case 'DICTIONARY':
                self.dictionary_init(limit, key)
            case 'REDIS':
                self.redis_init(limit, host, port, key)
            case _:
                raise BadStoreOption()
        

    def dictionary_init(self, limit: RateLimit, key: Optional[str]):
        self.store = DictionaryStore(limit, key)

    def redis_init(self, limit: RateLimit, host: str, port: int, key: Optional[str]):
        self.store = RedisStore(limit, host, port, key)

    def __call__(self, func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            self.store.update()
            func(*args, **kwargs)
        
        return wrapper
        





from .abstract_store import Store
from .exceptions import RateLimitedException 
from .rate_limit import RateLimit
import random
from time import monotonic

from typing import Optional


class DictionaryStore(Store):
    def __init__(self, limit: RateLimit, key: Optional[str]):
        self.limit = limit
        if key is None:
            self.key = str(random.randint(100000, 1000000))
        else:
            self.key = key
        self.d = {}

        self.set_tat(0)


    def get_tat(self):
        return self.d[self.key]

    def set_tat(self, value: str):
        self.d[self.key] = value

    def get_time(self):
        return monotonic()

    def update(self):
        now = self.get_time()
        tat = max(self.get_tat(), now)
        separation = tat - now
        max_interval = self.limit.period - self.limit.inverse
        if separation > max_interval:
            raise RateLimitedException()
        else:
            new_tat = max(tat, now) + self.limit.inverse
            self.set_tat(new_tat)



from redis import Redis 
from .abstract_store import Store 
from .rate_limit import RateLimit
from .exceptions import RateLimitedException
import random

from typing import Optional

LOCK_PREFIX = 'rate_limiter_lock:'

class RedisStore(Store):
    def __init__(
        self,
        limit: RateLimit,
        host: str,
        port: int,
        key: Optional[str],
    ):
        self.limit = limit
        self.r = Redis(host, port)
        if key is None:
            self.key = str(random.randint(100000, 1000000))
        else:
            self.key = key
        print(self.key)

    def get_tat(self):
        return float(self.r.get(self.key))

    def set_tat(self, value: str):
        return self.r.set(self.key, value)

    def get_time(self):
        t = self.r.time()
        return t[0] + t[1] / (10**6)

    def update(self):
        self.r.setnx(self.key, 0)
        now = self.get_time()

        with self.r.lock(f"{LOCK_PREFIX}{self.key}"):
            tat = max(self.get_tat(), now)
            separation = tat - now
            max_interval = self.limit.period - self.limit.inverse
            if separation > max_interval:
                raise RateLimitedException()
            else:
                new_tat = max(tat, now) + self.limit.inverse
                self.set_tat(new_tat)
